@@ -1,6 +1,7 @@
-/* Copyright (c) 1997-2015
-   Ewgenij Gawrilow, Michael Joswig (Technische Universitaet Berlin, Germany)
-   http://www.polymake.org
+/* Copyright (c) 1997-2023
+   Ewgenij Gawrilow, Michael Joswig, and the polymake team
+   Technische Universität Berlin, Germany
+   https://polymake.org
 
    This program is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
@@ -19,7 +20,7 @@
 
 namespace polymake { namespace topaz {
   
-void is_pseudo_manifold_client(perl::Object p)
+void is_pseudo_manifold_client(BigObject p)
 {
    const Lattice<BasicDecoration>& HD = p.give("HASSE_DIAGRAM");
    p.take("PSEUDO_MANIFOLD") << is_pseudo_manifold(HD,true);
